import string

from xcept import errors


class Formatter(string.Formatter):
    """Formatter for generating exception messages."""

    def __init__(self, allow_unused_args: bool):

        super().__init__()
        self._allow_unused_args = allow_unused_args

    def parse(self, format_string):

        for literal_text, field_name, format_spec, conversion in super().parse(format_string=format_string):
            if (field_name and (field_name in string.digits)) or (field_name == ""):
                raise errors.UsedPositionalArgumentError(f"positional argument is used!") from None
        iterator = super().parse(format_string=format_string)

        return iterator

    def check_unused_args(self, used_args, args, kwargs) -> None:

        if not self._allow_unused_args:
            for kwarg in kwargs:
                if kwarg not in used_args:
                    raise errors.UnusedKeywordArgumentError(
                        f"keyword argument {kwarg!r} is not used!", argument=kwarg
                    ) from None





class BaseError(Exception):

    def __init__(self, message):

        self.message = message

    def __str__(self):

        return self.message


class UsedPositionalArgumentError(BaseError):

    pass


class UnusedKeywordArgumentError(BaseError):

    def __init__(self, message, argument):

        super().__init__(message=message)
        self.argument = argument


class ArgsMatchingError(BaseError):

    def __init__(self, message, arg):

        super().__init__(message=message)
        self.arg = arg



from dataclasses import dataclass, field

from xcept.formatter import Formatter
from xcept import errors


_NON_FORMATTING_ATTRS = {"template_", "_formatter"}


@dataclass
class BaseException_(Exception):
    """Base class for exceptions."""

    ALLOW_UNUSED_ARGS = False

    template_: str = field(repr=False)

    def __post_init__(self):

        self._formatter = Formatter(allow_unused_args=self.ALLOW_UNUSED_ARGS)
        self._check_args_matching()

    def __str__(self):

        return self._get_message()

    @property
    def _formatting_attrs(self) -> dict:
        """Get attributes for formatting."""

        attrs = {}
        for key, value in vars(self).items():
            if key not in _NON_FORMATTING_ATTRS:
                attrs[key] = value

        return attrs

    def _get_message(self) -> str:
        """Get exception message."""

        return self._formatter.format(self.template_, **self._formatting_attrs)

    def _check_args_matching(self) -> None:
        """Check template arguments match attributes."""

        try:
            self._get_message()
        except KeyError as error:
            attr = error.args[0]
            raise errors.ArgsMatchingError(f"attribute {attr!r} specified in template "
                                           "was not found among the attributes!", arg=attr) from None



from dataclasses import dataclass

import pytest

from xcept import BaseException_
from xcept import errors


@pytest.mark.parametrize(
    ("template",),
    (
        ("err!",),
        ("error!",)
    )
)
def test(template):

    @dataclass
    class Exception_(BaseException_):
        pass

    with pytest.raises(Exception_, match=template):
        raise Exception_(template)


def test_with_args():

    @dataclass
    class Exception_(BaseException_):
        foo: str
        bar: int

    with pytest.raises(Exception_, match="error 'test' 123!"):
        raise Exception_("error {foo!r} {bar!s}!", foo="test", bar=123)


@pytest.mark.parametrize(
    ("template",),
    (
        ("{}",),
        ("{0}",),
        ("{}{0}",)
    )
)
def test_positional_arg_specified(template):

    @dataclass
    class Exception_(BaseException_):
        pass

    with pytest.raises(errors.UsedPositionalArgumentError):
        Exception_(template)


def test_keyword_arg_not_specified():

    @dataclass
    class Exception_(BaseException_):
        foo: str
        bar: int

    with pytest.raises(errors.UnusedKeywordArgumentError):
        Exception_("{foo}", foo="test", bar=123)


def test_args_matching():

    @dataclass
    class Exception_(BaseException_):
        foo: str

    with pytest.raises(errors.ArgsMatchingError):
        raise Exception_("error {foo} {unknown_arg}!", foo="test")


def test_unused_args_allowed():

    @dataclass
    class Exception_(BaseException_):
        ALLOW_UNUSED_ARGS = True
        foo: str
        bar: int

    with pytest.raises(Exception_, match="error test!"):
        raise Exception_("error {foo}!", foo="test", bar=123)
